unit EnergyNN;

interface

uses
  SysUtils, Math, Classes;

const
  INPUT_NEURONS = 4;  //   
  HIDDEN_NEURONS = 5;  //   
  OUTPUT_NEURONS = 1;   //   
 // EPOCHS = 1000;        //   
  LEARNING_RATE = 0.01; //  

type
DynAR2 = array of array of Double;
  TNeuron = class
  private
    FWeights: array of Double;
    FOutput: Double;
    FDelta: Double;
    function Sigmoid(x: Double): Double;
    function SigmoidDerivative(x: Double): Double;
  public
    constructor Create(InputCount: Integer);
    function FeedForward(Input: array of Double): Double;
    procedure CalculateDelta(Expected: Double);
    procedure UpdateWeights(Input: array of Double);
    property Output: Double read FOutput;
  end;

  TNeuralNetwork = class
  private
    FHiddenLayer: array of TNeuron;
    FOutputNeuron: TNeuron;
  public
    constructor Create;
    function Train(Input:DynAR2; Expected: array of Double;EPOCHS:Integer): Double;
    function Predict(Input: array of Double): Double;
  end;

implementation

{ TNeuron }

constructor TNeuron.Create(InputCount: Integer);
var
  i: Integer;
begin
  SetLength(FWeights, InputCount);
  Randomize;
  for i := 0 to InputCount - 1 do
    FWeights[i] := Random;  //   
end;

function TNeuron.Sigmoid(x: Double): Double;
begin
  Result := 1 / (1 + Exp(-x));
end;

function TNeuron.SigmoidDerivative(x: Double): Double;
begin
  Result := x * (1 - x);
end;

function TNeuron.FeedForward(Input: array of Double): Double;
var
  i: Integer;
  Sum: Double;
begin
  Sum := 0;
  for i := 0 to Length(FWeights) - 1 do
    Sum := Sum + (Input[i] * FWeights[i]);

  FOutput := Sigmoid(Sum);
  Result := FOutput;
end;

procedure TNeuron.CalculateDelta(Expected: Double);
begin
  FDelta := Expected - FOutput;
end;

procedure TNeuron.UpdateWeights(Input: array of Double);
var
  i: Integer;
begin
  for i := 0 to Length(FWeights) - 1 do
    FWeights[i] := FWeights[i] + LEARNING_RATE * FDelta * SigmoidDerivative(FOutput) * Input[i];
end;

{ TNeuralNetwork }

constructor TNeuralNetwork.Create;
var
  i: Integer;
begin
  SetLength(FHiddenLayer, HIDDEN_NEURONS);
  for i := 0 to HIDDEN_NEURONS - 1 do
    FHiddenLayer[i] := TNeuron.Create(INPUT_NEURONS);
  FOutputNeuron := TNeuron.Create(HIDDEN_NEURONS);
end;

function TNeuralNetwork.Train(Input: DynAR2; Expected: array of Double; EPOCHS:Integer): Double;
var
  HiddenOutputs: array of Double;
  i, j : Integer;
  TotalError: Double;
begin
  TotalError := 0;

  for j := 0 to EPOCHS - 1 do
  begin
    TotalError := 0;
    for i := 0 to High(Input) do
    begin
      //  
      for var k := 0 to HIDDEN_NEURONS - 1 do
        HiddenOutputs[k] := FHiddenLayer[k].FeedForward(Input[i]);

      var OutputPrediction := FOutputNeuron.FeedForward(HiddenOutputs);
      TotalError := TotalError + Sqr(Expected[i] - OutputPrediction);

      //  
      FOutputNeuron.CalculateDelta(Expected[i]);
      for var kk := 0 to HIDDEN_NEURONS - 1 do
        FHiddenLayer[kk].CalculateDelta(FOutputNeuron.Output * FOutputNeuron.FWeights[kk]);

      //  
      FOutputNeuron.UpdateWeights(HiddenOutputs);
      for var kkk := 0 to HIDDEN_NEURONS - 1 do
        FHiddenLayer[kkk].UpdateWeights(Input[i]);
    end;

    //  
    TotalError := TotalError / Length(Input);
  end;

  Result := TotalError;
end;

function TNeuralNetwork.Predict(Input: array of Double): Double;
var
  HiddenOutputs: array of Double;
  i: Integer;
begin
  SetLength(HiddenOutputs, HIDDEN_NEURONS);

  //  
  for i := 0 to HIDDEN_NEURONS - 1 do
    HiddenOutputs[i] := FHiddenLayer[i].FeedForward(Input);

  Result := FOutputNeuron.FeedForward(HiddenOutputs);
end;

end.
``
